////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2018 Josef Spidlen, Ph.D., FlowJo, LLC
//
// License
// The software is distributed under the terms of the 
// Artistic License 2.0
// http://www.r-project.org/Licenses/Artistic-2.0
// 
// Disclaimer
// This software and documentation come with no warranties of any kind.
// This software is provided "as is" and any express or implied 
// warranties, including, but not limited to, the implied warranties of
// merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the  copyright holder be liable for any direct, 
// indirect, incidental, special, exemplary, or consequential damages
// (including but not limited to, procurement of substitute goods or 
// services; loss of use, data or profits; or business interruption)
// however caused and on any theory of liability, whether in contract,
// strict liability, or tort arising in any way out of the use of this 
// software.    
//////////////////////////////////////////////////////////////////////////////

package com.flowjo.plugin.alra;
import java.awt.Component;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Box;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;

import com.flowjo.lib.parameters.ParameterSelectionPanel;
import com.flowjo.lib.parameters.ParameterSetMgrInterface;
import com.flowjo.lib.parameters.ParameterSelectionPanel.eParameterSelectionMode;
import com.treestar.lib.FJPluginHelper;
import com.treestar.lib.PluginHelper;
import com.treestar.lib.core.ExportFileTypes;
import com.treestar.lib.core.ExternalAlgorithmResults;
import com.treestar.lib.core.PopulationPluginInterface;
import com.treestar.lib.core.SeqGeqExternalAlgorithmResults;
import com.treestar.lib.data.StringUtil;
import com.treestar.lib.gui.FJButton;
import com.treestar.lib.gui.FontUtil;
import com.treestar.lib.gui.GuiFactory;
import com.treestar.lib.gui.HBox;
import com.treestar.lib.gui.actions.FJAction;
import com.treestar.lib.gui.numberfields.RangedIntegerTextField;
import com.treestar.lib.gui.panels.FJLabel;
import com.treestar.lib.gui.swing.FJCheckBox;
import com.treestar.lib.xml.SElement;

public class ALRA implements PopulationPluginInterface, ActionListener {

	public static final String pluginVersion = "0.2";
	public static final String pluginName = "ALRA";
	
	public static final int fixedToolTipW = 400;
	public static final int fixedLabelW = 10;
	public static final int fixedLabelW2 = 400;
	public static final int fixedLabelH = 25;
	public static final int fixedFieldW = 75;
	public static final int fixedFieldH = 25;

	public static final int parSelPaneW = 400;
	public static final int parSelPaneH = 300;

	public static final String Failed = "Failed";
	public static final String NoRepeat = "Cannot repeat";
	public static final String cellIdParName = "CellId";
	
	protected static final String fShowOutputButtonLabel = "Show ALRA R script output";
	protected static final String fShowOutputButtonTooltip = "Show the execution of the ALRA R script";
	
    public static final String explainLabel    = "Select parameters for Adaptively-thresholded Low-Rank Approximation (ALRA)";
    public static final String explainLabel2   = "Parameters used for ALRA.";
    public static final String explainTooltip  = "Single cell RNA-seq data is typically affected by so called dropouts - excess of zero or near zero counts due to the low amounts of mRNA within each individual cell. The ALRA plugin can impute the dropout values to improve the identification of cell types and further downsteam analysis.<br/>This is a selection of genes used for ALRA.";
    public static final String citationLabel   = "If you are using the ALRA plugin, please cite: <br/><font color=\"blue\">George C. Linderman, Jun Zhao, Yuval Kluger. Zero-preserving imputation of scRNA-seq data using low-rank approximation. bioRxiv. https://doi.org/10.1101/397588</font>";
    public static final String reviewOutLabel  = "<br/>The plugin execution typically results in the creation of a new imputed data set, which is automatically added to your workspace. Here, you can review the execution of the underlying R script that performed the imputation.";
    
	protected RangedIntegerTextField fKField = new RangedIntegerTextField(0, 200);
    public static final String rankLabel = "K";
    public static final String rankTooltip = "The rank of the rank-k approximation, use 0 to auto-estimate.";

	protected RangedIntegerTextField fQField = new RangedIntegerTextField(0, 20);
    public static final String qLabel = "Q";
    public static final String qTooltip = "The number of additional power iterations in randomized singular value decomposition; the default of 10 should be sufficient.";
    
    public static final String normalizeLabel = "Normalize";
    public static final String normalizeTooltip = "Normalize the data before ALRA imputation? Typically, you should keep this checked unless the data has been normalized already";
    protected FJCheckBox fNormalizeCheckbox = new FJCheckBox(normalizeLabel);
    
    private static final String htmlLabelPref   = "<html><p width=\"" + fixedLabelW2 + "\">";
    private static final String htmlLabelSuff   = "</p></html>";
    private static final String htmlTooltipPref = "<html><p width=\"" + fixedToolTipW + "\">";
    private static final String htmlTooltipSuff = "</p></html>";
    
	protected static Icon myIcon = null;
	protected static final String sIconName = "images/impIcon.png";
	
	protected List<String> fParameterNames = new ArrayList<String>();
	
	protected int fK = 0;
	protected int fQ = 10;
	protected boolean fNormalize = true;
	protected boolean fRanBefore = false;  // We only allow this plugin to run once, no recal, delete the node and try again if needed.
	protected String fRoutFile = "";
		
    @Override
	public String getName() 
	{
		return(pluginName);
	}

    @Override
	public String getVersion() 
	{
		return(pluginVersion);
	}
	
	@Override
	public Icon getIcon()
	{
		if (myIcon == null)
		{
			URL url = getClass().getClassLoader().getResource(getIconName());
			if (url != null)
				myIcon = new ImageIcon(url);
		}
		return myIcon;
	}
	
	protected String getIconName() { return sIconName; }
		
	@Override
	public ExportFileTypes useExportFileType() {
		return ExportFileTypes.CSV_SCALE;
	}

	public ExternalAlgorithmResults invokeAlgorithm(SElement fcmlQueryElement, File sampleFile, File outputFolder) 
	{
        SeqGeqExternalAlgorithmResults results = new SeqGeqExternalAlgorithmResults();
        String errMsg = "Something went wrong";
        
        if (fRanBefore) {
        	results.setWorkspaceString(ALRA.NoRepeat);
            return results;
        }
        fRanBefore = true;
        
        if(!sampleFile.exists()) 
        {
        	JOptionPane.showMessageDialog(null, "Input file did not exist", pluginName + " error", JOptionPane.ERROR_MESSAGE);
            results.setWorkspaceString(ALRA.Failed);
            return results;
        }
        
        if (fParameterNames.size() < 3)
		{
        	JOptionPane.showMessageDialog(null, "Not enough parameters were selected.", pluginName + " error", JOptionPane.ERROR_MESSAGE);
            results.setWorkspaceString(ALRA.Failed);
            return results;
		}
                
        String sampleName = sampleFile.getName();
		sampleName = sampleName.replaceAll(".FCS", "").replaceAll(".CSV", "").replaceAll(".ExtNode", "").replaceAll(".fcs", "").replaceAll(".LMD", "").replaceAll(".csv", "").trim();
        sampleName = StringUtil.rtrim(sampleName, ".");
        ALRACalc calculator = new ALRACalc();
        
        File resultCSV = calculator.runCalc(sampleFile, sampleName, fParameterNames, fK, fQ, fNormalize, outputFolder.getAbsolutePath());
        if (resultCSV == null) {
        	JOptionPane.showMessageDialog(null, errMsg, getName() + " error", JOptionPane.ERROR_MESSAGE);
            results.setWorkspaceString(ALRA.Failed);
        	return results;
        }
        
		File rErrFile = new File(resultCSV.getAbsolutePath().toString() + "_err.txt");
		if (rErrFile.exists()) {
			// Something went wrong, there is an error message in the file.
			try {
				errMsg = new String(Files.readAllBytes(Paths.get(rErrFile.getAbsolutePath())));
			} catch (IOException e) {
				e.printStackTrace();
			}
        	JOptionPane.showMessageDialog(null, errMsg, getName() + " error", JOptionPane.ERROR_MESSAGE);
        	results.setWorkspaceString(ALRA.Failed);
			rErrFile.delete();
			return results;
		}
        
        if (resultCSV != null && resultCSV.exists()) {
        	try {
        		File parentDir = resultCSV.getParentFile().getParentFile();
        		File movedResults = new File(parentDir, resultCSV.getName());
        		String[] fileNameAndExtension = parseFileName(resultCSV.getName());
        		int i = 1;
        		while (movedResults.exists() && i < 1000) {
        			movedResults = new File(parentDir, fileNameAndExtension[0] + " (" + i + ")" + fileNameAndExtension[1]);
        			i++;
        		}
        		if (movedResults.exists()) {
        			JOptionPane.showMessageDialog(null, "Failed to rename imputed results CSV file; file already exists.", getName() + " error", JOptionPane.ERROR_MESSAGE);
                    results.setWorkspaceString(ALRA.Failed);
                    return results;
        		} else {
        			resultCSV.renameTo(movedResults);
        			FJPluginHelper.loadSamplesIntoWorkspace(fcmlQueryElement, new String[] { movedResults.getAbsolutePath().toString() } );
        		}
        	} catch (Exception e) {
        		JOptionPane.showMessageDialog(null, "Failed to move imputed results CSV file to the workspace directory.", getName() + " error", JOptionPane.ERROR_MESSAGE);
                results.setWorkspaceString(ALRA.Failed);
                return results;
        	}        	
        } else {
        	JOptionPane.showMessageDialog(null, "ALRA R script failed to generate the results", getName() + " error", JOptionPane.ERROR_MESSAGE);
            results.setWorkspaceString(ALRA.Failed);
            return results;
        }
        calculator.deleteScriptFile();
        
		File rOutFile = calculator.renameROutFile();
		if (rOutFile != null)
			fRoutFile = rOutFile.getAbsolutePath();

        results.setWorkspaceString("DONE!");
        return results;
	}

	private String[] parseFileName(String name)
	{
		String[] res = new String[2];
		res[0] = name.replaceFirst("[.][^.]+$", "");
		res[1] = "";
		int i = name.lastIndexOf('.');
		if (i > 0)
			res[1] = name.substring(i);
		return res;
	}

	@Override
	public SElement getElement()
	{
		SElement result = new SElement(getName());
		if (!fParameterNames.isEmpty())
		{
			SElement elem = new SElement(XMLConsts.Parameters);
			result.addContent(elem);
			for (String pName : fParameterNames)
			{
				SElement e = new SElement(XMLConsts.P);
				e.setString(XMLConsts.name, pName);
				elem.addContent(e);
			}
		}
		result.setInt(XMLConsts.K, fK);
		result.setInt(XMLConsts.Q, fQ);
		result.setBool(XMLConsts.normalize, fNormalize);
		result.setBool(XMLConsts.ranBefore, fRanBefore);
		result.setString(XMLConsts.routFile, fRoutFile);
		
		return result;	
	}

	@Override
	public void setElement(SElement element)
	{
		SElement params = element.getChild(XMLConsts.Parameters);
		if (params != null) {
			fParameterNames.clear();
			for (SElement elem : params.getChildren())
			{
				fParameterNames.add(elem.getString(XMLConsts.name));
			}			
		}
		fK = element.getInt(XMLConsts.K, fK);
		fQ = element.getInt(XMLConsts.Q, fQ);
		fNormalize = element.getBool(XMLConsts.normalize, fNormalize);
		fRanBefore = element.getBool(XMLConsts.ranBefore, fRanBefore);
		fRoutFile = element.getString(XMLConsts.routFile, fRoutFile);
	}

	@Override
	public List<String> getParameters()
	{
		return fParameterNames;
	}

	@Override
	public boolean promptForOptions(SElement fcmlQueryElement, List<String> pNames)
	{
		ParameterSetMgrInterface mgr = PluginHelper.getParameterSetMgr(fcmlQueryElement);
		if (mgr == null)
			return false;		
		
		List<Object> guiObjects = new ArrayList<Object>();
		
		FJLabel explainText = new FJLabel(explainLabel);
		if (fRanBefore) explainText.setText(explainLabel2);
		explainText.setToolTipText(htmlTooltipPref + explainTooltip + htmlTooltipSuff);
		guiObjects.add(new HBox(new Component[] { explainText }));
		ParameterSelectionPanel parSelPanel = null;
		
		if (fRanBefore) {			
			JList<String> jlistScroll = new JList<String>((String[])fParameterNames.toArray(new String[0]));
		    jlistScroll.setEnabled(false);
		    JScrollPane sp = new JScrollPane(jlistScroll);
		    Dimension dim = new Dimension(parSelPaneW, parSelPaneH / 2);
		    sp.setMaximumSize(dim);
			sp.setMinimumSize(dim);
			sp.setPreferredSize(dim);
			jlistScroll.setToolTipText(htmlTooltipPref + explainTooltip + htmlTooltipSuff);
			sp.setToolTipText(htmlTooltipPref + explainTooltip + htmlTooltipSuff);
		    guiObjects.add(sp);
		} else {
			parSelPanel = new ParameterSelectionPanel(mgr, eParameterSelectionMode.WithSetsAndParameters, true, false, false, true);
			parSelPanel.setSelectedParameters(fParameterNames);
			Dimension dim = new Dimension(parSelPaneW, parSelPaneH);
			parSelPanel.setMaximumSize(dim);
			parSelPanel.setMinimumSize(dim);
			parSelPanel.setPreferredSize(dim);
			guiObjects.add(parSelPanel);
		}
		
        FJLabel fjLabelK = new FJLabel(rankLabel);
        fKField = new RangedIntegerTextField(0, 200);
        fKField.setInt(fK);
        fKField.setToolTipText(htmlTooltipPref + rankTooltip + htmlTooltipSuff);
        
        GuiFactory.setSizes(fKField, new Dimension(fixedFieldW, fixedFieldH));
        GuiFactory.setSizes(fjLabelK, new Dimension(fixedLabelW, fixedLabelH));

        FJLabel fjLabelQ = new FJLabel(qLabel);
        fQField = new RangedIntegerTextField(0, 20);
        fQField.setInt(fQ);
        fQField.setToolTipText(htmlTooltipPref + qTooltip + htmlTooltipSuff);

        GuiFactory.setSizes(fQField, new Dimension(fixedFieldW, fixedFieldH));
        GuiFactory.setSizes(fjLabelQ, new Dimension(fixedLabelW, fixedLabelH));

        fNormalizeCheckbox = new FJCheckBox(normalizeLabel);
        fNormalizeCheckbox.setToolTipText(htmlTooltipPref + normalizeTooltip + htmlTooltipSuff);
        fNormalizeCheckbox.setSelected(fNormalize);
        
        guiObjects.add(new HBox(new Component[] { fjLabelK, fKField, Box.createHorizontalGlue(), fjLabelQ, fQField, Box.createHorizontalGlue(), fNormalizeCheckbox }));
        
		FJLabel citationText = new FJLabel(htmlLabelPref + citationLabel + htmlLabelSuff);
		citationText.setToolTipText(htmlTooltipPref + citationLabel + htmlTooltipSuff);
		guiObjects.add(new HBox(new Component[] { citationText }));
		
		if (fRanBefore) {
			fKField.setEnabled(false);
			fQField.setEnabled(false);
			fNormalizeCheckbox.setEnabled(false);
			
			if (fRoutFile != null && !fRoutFile.isEmpty()) {
				File rOut = new File(fRoutFile);
				if (rOut != null && rOut.exists()) {					
					FJLabel reviewOutText = new FJLabel(htmlLabelPref + reviewOutLabel + htmlLabelSuff);
					reviewOutText.setToolTipText(htmlTooltipPref + reviewOutLabel + htmlTooltipSuff);
					guiObjects.add(new HBox(new Component[] { reviewOutText }));
					guiObjects.add(new HBox(Box.createHorizontalGlue(),  createShowOutputButton(), Box.createHorizontalGlue()));
				}
			}			
			JOptionPane.showMessageDialog((Component) null, guiObjects.toArray(), getName() + " " + getVersion(), JOptionPane.PLAIN_MESSAGE);
			return false;
		} else {
			int option = JOptionPane.showConfirmDialog((Component) null, guiObjects.toArray(), getName() + " " + getVersion(), JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE);
			if (option == JOptionPane.OK_OPTION)
			{
				fK = fKField.getInt();
				fQ = fQField.getInt();
				fNormalize = fNormalizeCheckbox.isSelected();
				
				if(parSelPanel != null) {
					fParameterNames.clear();
					fParameterNames.addAll(parSelPanel.getParameterSelection());
				}
				return true;
			}
			else
				return false;
		}
	}

	private Component createShowOutputButton()
	{
		FJButton outButton = new FJButton(new FJAction(this, fShowOutputButtonLabel), FontUtil.dlog10);
		outButton.setToolTipText(fShowOutputButtonTooltip);
		return outButton;
	}	

	@Override
	public void actionPerformed(ActionEvent e)
	{
		try {
			if(fShowOutputButtonLabel.equals(e.getActionCommand()) && fRoutFile != null && !fRoutFile.isEmpty())
			{
				File file = new File(fRoutFile);
				if (file != null && file.exists())
					Desktop.getDesktop().open(file);
				else
					Desktop.getDesktop().browse(new URI(fRoutFile));
			}
		}
		catch (IOException e1) { }
		catch (URISyntaxException e1) { }
	}
}
