#######################################################################
# Copyright (c) 2018 Josef Spidlen, Ph.D., FlowJo, LLC
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################
##
## The core functionality has been adapted from
## Adapted from https://github.com/KlugerLab/ALRA
##
######################################################################

suppressMessages(library("rsvd")) # Tested with rsvd_0.9
sessionInfo()

# cat("FJ-DATA-FILE-PATH         = FJ_DATA_FILE_PATH \n")
# cat("FJ-CSV-OUPUT-FILE         = FJ_CSV_OUPUT_FILE \n")
# cat("FJ-PAR-K                  = FJ_PAR_K \n")
# cat("FJ-PAR-Q                  = FJ_PAR_Q \n")
# cat("FJ-PAR-NORMALIZE          = FJ_PAR_NORMALIZE \n")
# cat("FJ-PARAMS-LIST            = FJ-PARAMS-LIST \n")     ## This is too long, we don't want this... that's why "-" instead of "_" 

outCSVPath <- "FJ_CSV_OUPUT_FILE"
inCSVPath <- "FJ_DATA_FILE_PATH"

fj_K <- as.numeric(FJ_PAR_K)
fj_Q <- as.numeric(FJ_PAR_Q)
fj_Normalize <- FJ_PAR_NORMALIZE

stopAndReturnError <- function(message, ...) {
	cat(message, file=paste0(outCSVPath, "_err.txt"))
	stop(message, ...)
}

library(rsvd)

normalize_data <- function (A) {
	#  Simple convenience function to library and log normalize a matrix
	totalUMIPerCell <- rowSums(A);
	A_norm <- sweep(A, 1, totalUMIPerCell, '/');
	A_norm <- A_norm[complete.cases(A_norm), ]
	A_norm <- A_norm * 10E3
	A_norm <- log(A_norm + 1);
	if (any(totalUMIPerCell == 0)) {
		A_norm <- A_norm[-which(totalUMIPerCell == 0),]
	}
	A_norm
}

choose_k <- function (A_norm, K=100, pval_thresh=1E-10, noise_start=80) {
	#  Heuristic for choosing rank k for the low rank approximation based on
	#  statistics of the spacings between consecutive singular values. Finds
	#  the smallest singular value \sigma_i such that $\sigma_i - \sigma_{i-1}
	#  is significantly different than spacings in the tail of the singular values.
	# 
	#
	# Args:
	#   A_norm: The log-transformed expression matrix of cells (rows) vs. genes (columns)
	#   pval_thresh : The threshold for ``significance''
	#   noise_start : Index for which all smaller singular values are considered noise
	#
	# Returns:
	#   A list with three items
	#       1) Chosen k
	#       2) P values of each possible k 
	#       3) Singular values of the matrix A_norm
	
	if (K > min(dim(A_norm))) {
		stop("For an m by n matrix, K must be smaller than the min(m,n).")
	}
	if (noise_start >K-5) {
		stop("There need to be at least 5 singular values considered noise.")
	}
	
	noise_svals <- noise_start:K
	rsvd_out <- rsvd(A_norm,K)
	diffs <- diff(rsvd_out$d)
	pvals <- pnorm(diffs,mean(diffs[noise_svals-1]), sd(diffs[noise_svals-1]))
	k <- max(which(pvals < pval_thresh))
	return (list(k=k, pvals=pvals, d=rsvd_out$d))
}

alra <- function(A_norm, k=0, q=10) {
	# Computes the k-rank approximation to A_norm and adjusts it according to the
	# error distribution learned from the negative values.
	#
	# Args:
	#   A_norm: The log-transformed expression matrix of cells (rows) vs. genes (columns)
	#   k : the rank of the rank-k approximation. Set to 0 for automated choice of k.
	#   q : the number of additional power iterations in randomized SVD
	#
	# Returns:
	#   A list with three items
	#       1) The rank k approximation of A_norm.
	#       2) The rank k approximation of A_norm, adaptively thresholded
	#       3) The rank k approximation of A_norm, adaptively thresholded and with the first two moments of the non-zero values matched to the first two moments of the non-zeros of A_norm
	# Example:
	#     result.completed <- adjusted_svd(A_norm,15)
	#     A_norm_rank15 <- result.completed[[1]]     # The low rank approximation for reference purposes...not suggested for matrix completion
	#     A_norm_rank15_cor <- result.completed[[3]] # The actual adjusted, completed matrix
	if (k == 0 ) {
		
		if (min(dim(A_norm)) < 10) {
			stopAndReturnError("The dimensionality of the input is too small, you need at least 10 cells and 10 genes to estimate good rank for ALRA.")
		}
		
		K <- 100
		noise_start <- 80
		if (K >= min(dim(A_norm))) {
			K <- min(dim(A_norm)) - 1
		}
		if (noise_start > K-5) {
			if (K >= 30) {
				noise_start <- as.integer(K*0.8) 
			} else {
				noise_start <- K-6
			}
		}
		
		k_choice <- choose_k(A_norm, K=K, noise_start=noise_start)
		k <- k_choice$k
		if (k == -Inf) {
			stopAndReturnError("Failed to set K automatically.")
		}
		print(sprintf("Chose k=%d",k))
	}
	
	print("Getting nonzeros")
	originally_nonzero <- A_norm >0 
	
	print("Randomized SVD")
	fastDecomp_noc <- rsvd(A_norm, k,q=q);
	A_norm_rank_k <- fastDecomp_noc$u[,1:k]%*%diag(fastDecomp_noc$d[1:k])%*% t(fastDecomp_noc$v[,1:k])
	
	
	print("Find mins")
	A_norm_rank_k_mins <- abs(apply(A_norm_rank_k,2,min))
	print("Sweep")
	A_norm_rank_k_cor <- replace(A_norm_rank_k, A_norm_rank_k <= A_norm_rank_k_mins[col(A_norm_rank_k)], 0)
	
	
	sd_nonzero <- function(x) sd(x[!x == 0])
	sigma_1 <- apply(A_norm_rank_k_cor, 2, sd_nonzero)
	sigma_2 <- apply(A_norm, 2, sd_nonzero)
	mu_1 <- colSums(A_norm_rank_k_cor)/colSums(!!A_norm_rank_k_cor)
	mu_2 <- colSums(A_norm)/colSums(!!A_norm)
	
	toscale <- !is.na(sigma_1) & !is.na(sigma_2)
	
	print(sprintf("Scaling all except for %d columns", sum(!toscale)))
	
	sigma_1_2 <- sigma_2/sigma_1
	toadd  <- -1*mu_1*sigma_2/sigma_1 + mu_2
	
	A_norm_rank_k_temp <- A_norm_rank_k_cor[,toscale]
	A_norm_rank_k_temp <- sweep(A_norm_rank_k_temp,2, sigma_1_2[toscale],FUN = "*")
	A_norm_rank_k_temp <- sweep(A_norm_rank_k_temp,2, toadd[toscale],FUN = "+")
	
	A_norm_rank_k_cor_sc <- A_norm_rank_k_cor
	A_norm_rank_k_cor_sc[,toscale] <- A_norm_rank_k_temp
	A_norm_rank_k_cor_sc[A_norm_rank_k_cor==0] = 0
	
	lt0 <- A_norm_rank_k_cor_sc  <0
	A_norm_rank_k_cor_sc[lt0] <- 0 
	print(sprintf("%.2f%% of the values became negative in the scaling process and were set to zero", 100*sum(lt0)/(nrow(A_norm)*ncol(A_norm))))
	
	A_norm_rank_k_cor_sc[originally_nonzero & A_norm_rank_k_cor_sc ==0] <- A_norm[originally_nonzero & A_norm_rank_k_cor_sc ==0]
	
	colnames(A_norm_rank_k_cor) <- colnames(A_norm)
	colnames(A_norm_rank_k_cor_sc) <- colnames(A_norm)
	colnames(A_norm_rank_k) <- colnames(A_norm)
	
	original_nz <- sum(A_norm>0)/(nrow(A_norm)*ncol(A_norm))
	completed_nz <- sum(A_norm_rank_k_cor_sc>0)/(nrow(A_norm)*ncol(A_norm))
	print(sprintf("The matrix went from %.2f%% nonzero to %.2f%% nonzero", 100*original_nz, 100*completed_nz))
	
	list(A_norm_rank_k=A_norm_rank_k,A_norm_rank_k_cor =A_norm_rank_k_cor, A_norm_rank_k_cor_sc=A_norm_rank_k_cor_sc)
}

A <- as.matrix(read.csv(inCSVPath))
if (fj_Normalize) {
	A <- normalize_data(A)
}

result <- tryCatch({
  	alra(A, k=fj_K, q=fj_Q)
  }, error = function(e) {
	stopAndReturnError("Failed to impute the data, try different input settings (e.g., different k value.)");
  }
)

result_cor <- result$A_norm_rank_k_cor
CellId <- as.matrix(1:dim(result_cor)[1], ncol=1)
colnames(CellId) <- 'CellId'
result_final <- cbind(CellId, result_cor)

write.csv(result_final, file=outCSVPath, row.names=FALSE)

# Clean up in case R was saving the .RData when exiting
rm(list=ls())
